//#####################################################################
// Copyright 2003-2010, Ronald Fedkiw, Geoffrey Irving, Nipun Kwatra, Michael Lentine, Duc Nguyen, Andrew Selle.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class CUBIC_MN_INTERPOLATION_UNIFORM 
//#####################################################################
#ifndef __CUBIC_MN_INTERPOLATION_UNIFORM__
#define __CUBIC_MN_INTERPOLATION_UNIFORM__

#include <PhysBAM_Tools/Grids_Uniform_Interpolation/INTERPOLATION_UNIFORM.h>
#include <PhysBAM_Tools/Interpolation/CUBIC_MN_INTERPOLATION.h>
namespace PhysBAM{

template<class T_GRID,class T2,class T_FACE_LOOKUP> // T_FACE_LOOKUP=FACE_LOOKUP_UNIFORM<T_GRID>
class CUBIC_MN_INTERPOLATION_UNIFORM:public INTERPOLATION_UNIFORM<T_GRID,T2,T_FACE_LOOKUP>
{
    typedef typename T_GRID::VECTOR_T TV;typedef typename TV::SCALAR T;typedef typename T_GRID::VECTOR_INT TV_INT;
    //typedef typename GRID_ARRAYS_POLICY<T_GRID>::ARRAYS_SCALAR::template REBIND<T2>::TYPE T_ARRAYS_T2;
    typedef ARRAYS_ND_BASE<VECTOR<T2,TV::dimension> > T_ARRAYS_T2;
public:
    typedef INTERPOLATION_UNIFORM<T_GRID,T2,T_FACE_LOOKUP> BASE;
    using BASE::Clamped_Index_Interior_End_Minus_One;

    T b,c; // two parameter family
private:
    CUBIC_MN_INTERPOLATION<T,T2> cubic_mn_interpolation;
public:

    CUBIC_MN_INTERPOLATION_UNIFORM()
    {
        Set_Sharpness();
    }

    void Set_Sharpness(T sharpness_input=2./3)
    {Set_Parameters(1-sharpness_input,(T).5*sharpness_input);}

    void Set_Parameters(const T b_input=1./3,const T c_input=1./3)
    {b=b_input;c=c_input;
    cubic_mn_interpolation.Set_Parameters(b,c);}

    T2 Clamped_To_Array(const T_GRID& grid,const T_ARRAYS_T2& u,const TV& X) const PHYSBAM_OVERRIDE
    {return From_Base_Node(grid,u,X,INTERPOLATION_UNIFORM<T_GRID,T2,T_FACE_LOOKUP>::Clamped_Index_Interior_End_Minus_One(grid,u,X)-TV_INT::All_Ones_Vector());}
    
    ARRAY<PAIR<TV_INT,T> > Clamped_To_Array_Weights(const T_GRID& grid,const T_ARRAYS_T2& u,const TV& X) const PHYSBAM_OVERRIDE
    {return From_Base_Node_Weights(grid,u,X,INTERPOLATION_UNIFORM<T_GRID,T2,T_FACE_LOOKUP>::Clamped_Index_Interior_End_Minus_One(grid,u,X)-TV_INT::All_Ones_Vector());}
    
    T2 Periodic(const T_GRID& grid,const T_ARRAYS_T2& u,const TV& X) const
    {return From_Base_Node_Periodic(grid,u,X,grid.Cell(X,0)-TV_INT::All_Ones_Vector());}
    
    T2 From_Base_Node(const GRID<TV>& grid,const ARRAYS_ND_BASE<VECTOR<T2,1> >& u,const VECTOR<T,1>& X,const VECTOR<int,1>& index) const
    {return cubic_mn_interpolation.Cubic_MN(u(index.x),u(index.x+1),u(index.x+2),u(index.x+3),(X.x-grid.X(index.x+1).x)*grid.one_over_dX.x);}
    
    ARRAY<PAIR<TV_INT,T> > From_Base_Node_Weights(const GRID<TV>& grid,const ARRAYS_ND_BASE<VECTOR<T2,1> >& u,const VECTOR<T,1>& X,const VECTOR<int,1>& index) const
    {ARRAY<T> local_weights;local_weights=cubic_mn_interpolation.Cubic_MN_Weights((X.x-grid.X(index.x+1).x)*grid.one_over_dX.x);
    ARRAY<PAIR<TV_INT,T> > weights;for(int i=0;i<4;i++) weights.Append(PAIR<TV_INT,T>(TV_INT(index.x+i),local_weights(i+1)));return weights;}
    
    T2 From_Base_Node(const GRID<TV>& grid,const ARRAYS_ND_BASE<VECTOR<T2,2> >& u,const VECTOR<T,2>& X,const VECTOR<int,2>& index) const
    {VECTOR<T,2> alpha=X-grid.X(index+VECTOR<int,2>::All_Ones_Vector());alpha*=grid.one_over_dX;
    T2 interpolated_in_x[4],value[4];
    for(int b=0;b<4;b++){
        for(int a=0;a<4;a++) value[a]=u(index.x+a,index.y+b);
        interpolated_in_x[b]=cubic_mn_interpolation.Cubic_MN(value[0],value[1],value[2],value[3],alpha.x);}
    return cubic_mn_interpolation.Cubic_MN(interpolated_in_x[0],interpolated_in_x[1],interpolated_in_x[2],interpolated_in_x[3],alpha.y);}

    ARRAY<PAIR<TV_INT,T> > From_Base_Node_Weights(const GRID<TV>& grid,const ARRAYS_ND_BASE<VECTOR<T2,2> >& u,const VECTOR<T,2>& X,const VECTOR<int,2>& index) const
    {VECTOR<T,2> alpha=X-grid.X(index+VECTOR<int,2>::All_Ones_Vector());alpha*=grid.one_over_dX;
    ARRAY<PAIR<TV_INT,T> > weights;
    ARRAY<T> local_weights_x=cubic_mn_interpolation.Cubic_MN_Weights(alpha.x);
    ARRAY<T> local_weights_y=cubic_mn_interpolation.Cubic_MN_Weights(alpha.y);
    for(int b=0;b<4;b++) for(int a=0;a<4;a++) weights.Append(PAIR<TV_INT,T>(TV_INT(index.x+a,index.y+b),local_weights_y(b+1)*local_weights_x(a+1)));
    return weights;}
    
    ARRAY<PAIR<TV_INT,T> > From_Base_Node_Weights(const GRID<TV>& grid,const ARRAYS_ND_BASE<VECTOR<T2,3> >& u,const VECTOR<T,3>& X,const VECTOR<int,3>& index) const
    {ARRAY<PAIR<TV_INT,T> > weights;
    PHYSBAM_NOT_IMPLEMENTED();
    return weights;}

    // only works for power of two grids!
    T2 From_Base_Node_Periodic(const GRID<TV>& grid,const ARRAYS_ND_BASE<VECTOR<T2,2> >& u,const VECTOR<T,2>& X,const VECTOR<int,2>& index) const
    {VECTOR<T,2> alpha=X-grid.X(index+VECTOR<int,2>::All_Ones_Vector());alpha*=grid.one_over_dX;
    T2 interpolated_in_x[4],value[4];
    for(int b=0;b<4;b++){
        for(int a=0;a<4;a++){
            int x=index.x+a-1,y=index.y+b-1;
            while(x<0) x+=grid.m;if(x>=grid.m)x=(x&(grid.m-1));
            while(y<0) y+=grid.n;if(y>=grid.n)y=(y&(grid.n-1));
            value[a]=u(x+1,y+1);}
        interpolated_in_x[b]=cubic_mn_interpolation.Cubic_MN(value[0],value[1],value[2],value[3],alpha.x);}
    return cubic_mn_interpolation.Cubic_MN(interpolated_in_x[0],interpolated_in_x[1],interpolated_in_x[2],interpolated_in_x[3],alpha.y);}

    T2 From_Base_Node(const GRID<TV>& grid,const ARRAYS_ND_BASE<VECTOR<T2,3> >& u,const VECTOR<T,3>& X,const VECTOR<int,3>& index) const
    {VECTOR<T,3> alpha=X-grid.X(index+VECTOR<int,3>::All_Ones_Vector());alpha*=grid.one_over_dX;
    T2 interpolated_in_x[4],interpolated_in_y[4],value[4];
    for(int c=0;c<4;c++){
        for(int b=0;b<4;b++){
            for(int a=0;a<4;a++) value[a]=u(index.x+a,index.y+b,index.z+c);
            interpolated_in_x[b]=cubic_mn_interpolation.Cubic_MN(value[0],value[1],value[2],value[3],alpha.x);}
        interpolated_in_y[c]=cubic_mn_interpolation.Cubic_MN(interpolated_in_x[0],interpolated_in_x[1],interpolated_in_x[2],interpolated_in_x[3],alpha.y);}
    return cubic_mn_interpolation.Cubic_MN(interpolated_in_y[0],interpolated_in_y[1],interpolated_in_y[2],interpolated_in_y[3],alpha.z);}

//#####################################################################
};
}
#endif
