//#####################################################################
// Copyright 2010 Michael Lentine.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class PCG_SPARSE_THREADED
//#####################################################################
#ifndef __PCG_SPARSE_THREADED__
#define __PCG_SPARSE_THREADED__

#include <PhysBAM_Tools/Grids_Uniform/GRID.h>
#include <PhysBAM_Tools/Grids_Uniform/UNIFORM_GRID_ITERATOR_CELL.h>
#include <PhysBAM_Tools/Krylov_Solvers/PCG_SPARSE.h>
#include <PhysBAM_Tools/Parallel_Computation/PTHREAD.h>
#include <PhysBAM_Tools/Parallel_Computation/THREAD_QUEUE.h>
namespace PhysBAM{

template<class TV>
class PCG_SPARSE_THREADED:public PCG_SPARSE<typename TV::SCALAR>
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::dimension> TV_INT;
    typedef typename GRID<TV>::CELL_ITERATOR CELL_ITERATOR;
    typedef PCG_SPARSE<T> BASE;
public:
    THREAD_QUEUE& thread_queue;
    bool sum_int_needs_init,sum_needs_init,max_needs_init,barr_needs_init;
    bool needs_sync;
    T sum,max_val;
    int sum_int,number_of_threads;
    VECTOR_ND<T> temp,p;
#ifdef USE_PTHREADS
    pthread_mutex_t barr_lock,sum_lock,max_lock,sum_int_lock;
    pthread_mutex_t lock;
    pthread_barrier_t barr;
#endif

    using BASE::remove_null_space_solution_component;using BASE::enforce_compatibility;using BASE::maximum_iterations;
    using BASE::show_results;using BASE::show_residual;
    using BASE::incomplete_cholesky;using BASE::modified_incomplete_cholesky;using BASE::modified_incomplete_cholesky_coefficient;
    using BASE::preconditioner_zero_tolerance;using BASE::preconditioner_zero_replacement;

    PCG_SPARSE_THREADED(THREAD_QUEUE* thread_queue_input)
        :thread_queue(*thread_queue_input),sum_int_needs_init(false),sum_needs_init(false),max_needs_init(false),barr_needs_init(true),needs_sync(true),sum(0),max_val(0),sum_int(0)
    {
#ifdef USE_PTHREADS
        pthread_mutex_init(&sum_lock,0);
        pthread_mutex_init(&sum_int_lock,0);
        pthread_mutex_init(&max_lock,0);
        pthread_mutex_init(&barr_lock,0);
        pthread_mutex_init(&lock,0);
#endif
    }

    ~PCG_SPARSE_THREADED()
    {}

    void Init_Barriers()
    {
#ifdef USE_PTHREADS
        pthread_mutex_lock(&barr_lock);
        if(barr_needs_init){
            pthread_barrier_destroy(&barr);
            pthread_barrier_init(&barr,0,number_of_threads);
            barr_needs_init=false;}
        pthread_mutex_unlock(&barr_lock);
        pthread_barrier_wait(&barr);
        barr_needs_init=true;
#endif
    }

    int Global_Sum_Int(int input)
    {
#ifdef USE_PTHREADS
        pthread_mutex_lock(&sum_int_lock);
        if(sum_int_needs_init){sum_int=input;sum_int_needs_init=false;}
        else sum_int+=input;
        pthread_mutex_unlock(&sum_int_lock);
        pthread_barrier_wait(&barr);
        sum_int_needs_init=true;
        return sum_int;
#else
        return input;
#endif
    }


    T Global_Sum(T input)
    {
#ifdef USE_PTHREADS
        pthread_mutex_lock(&sum_lock);
        // update this to be deterministic
        if(sum_needs_init){sum=input;sum_needs_init=false;}
        else sum+=input;
        pthread_mutex_unlock(&sum_lock);
        pthread_barrier_wait(&barr);
        sum_needs_init=true;
        return sum;
#else
        return input;
#endif
    }

    T Global_Max(T input)
    {
#ifdef USE_PTHREADS
        pthread_mutex_lock(&max_lock);
        if(max_needs_init){max_val=input;max_needs_init=false;}
        else max_val=max(max_val,input);
        pthread_mutex_unlock(&max_lock);
        pthread_barrier_wait(&barr);
        max_needs_init=true;
        return max_val;
#else
        return input;
#endif
    }

    void Solve(SPARSE_MATRIX_FLAT_NXN<T>& A_matrix,VECTOR_ND<T>& x,VECTOR_ND<T>& b,VECTOR_ND<T>& q,VECTOR_ND<T>& s,VECTOR_ND<T>& r,VECTOR_ND<T>& k,VECTOR_ND<T>& z,const T tolerance=1e-7,const bool recompute_preconditioner=true)
    {
        PHYSBAM_FATAL_ERROR();
    }

//#####################################################################
    void Solve(RANGE<TV_INT>& domain,const ARRAY<int,TV_INT>& domain_index,const ARRAY<INTERVAL<int> >& all_interior_indices,const ARRAY<ARRAY<INTERVAL<int> > >& all_ghost_indices,SPARSE_MATRIX_FLAT_NXN<T>& A,VECTOR_ND<T>& x,VECTOR_ND<T>& b,const T tolerance);
//#####################################################################
};
}
#endif
